import random
import time
import unicodedata
from dataclasses import field
from typing import ClassVar, Optional

import clabe
from pydantic import PositiveFloat, conint, constr, validator
from pydantic.dataclasses import dataclass

from ..auth import ORDEN_FIELDNAMES, compute_signature, join_fields
from ..types import Prioridad, TipoCuenta, digits, truncated_str
from .base import Resource

STP_BANK_CODE = '90646'


@dataclass
class Orden(Resource):
    """
    Base on:
    https://stpmex.zendesk.com/hc/es/articles/360002682851-RegistraOrden-Dispersi%C3%B3n-
    """

    _endpoint: ClassVar[str] = '/ordenPago'

    monto: PositiveFloat
    conceptoPago: truncated_str(39)

    nombreBeneficiario: truncated_str(39)
    cuentaBeneficiario: digits(10, 19)
    institucionContraparte: digits(5, 5)
    tipoCuentaBeneficiario: int

    cuentaOrdenante: digits(18, 18)

    nombreOrdenante: Optional[truncated_str(39)] = None
    institucionOperante: digits(5, 5) = STP_BANK_CODE
    tipoCuentaOrdenante: Optional[int] = None

    claveRastreo: truncated_str(29) = field(
        default_factory=lambda: f'CR{int(time.time())}'
    )
    referenciaNumerica: conint(gt=0, lt=10 ** 7) = field(
        default_factory=lambda: random.randint(10 ** 6, 10 ** 7)
    )
    rfcCurpBeneficiario: constr(max_length=18) = 'ND'
    rfcCurpOrdenante: Optional[constr(max_length=18)] = None
    medioEntrega: int = 3
    prioridad: int = Prioridad.alta.value
    tipoPago: int = 1
    topologia: str = 'T'
    iva: Optional[float] = None

    id: Optional[int] = None

    @classmethod
    def registra(cls, **kwargs):
        orden = cls(**kwargs)
        endpoint = orden._endpoint + '/registra'
        resp = orden._client.put(endpoint, orden.to_dict())
        orden.id = resp['id']
        return orden

    @property
    def firma(self):
        """
        Based on:
        https://stpmex.zendesk.com/hc/es/articles/360002796012-Firmas-Electr%C3%B3nicas-
        """
        joined_fields = join_fields(self, ORDEN_FIELDNAMES)
        return compute_signature(self._client.pkey, joined_fields)

    def __post_init__(self):
        # Test before Pydantic coerces it to a float
        if not isinstance(self.monto, float):
            raise ValueError('monto must be a float')

    @validator('cuentaBeneficiario', 'cuentaOrdenante', each_item=True)
    def _validate_cuenta(cls, v):
        if len(v) == 18:
            if not clabe.validate_clabe(v):
                raise ValueError('cuenta no es una válida CLABE')
        elif not len(v) in {10, 15, 16}:
            raise ValueError('cuenta no es válida')
        return v

    @validator('institucionContraparte', 'institucionOperante', each_item=True)
    def _validate_institucion(cls, v):
        if v not in clabe.BANKS.values():
            raise ValueError(f'{v} no se corresponde a un banco')
        return v

    @staticmethod
    def _validate_tipoCuenta(fieldname, tipo, values):
        try:
            cuenta = values[fieldname]
        except KeyError:  # there's a validation error elsewhere
            return tipo
        if not any(
            [
                len(cuenta) == 10 and tipo == TipoCuenta.phone_number.value,
                len(cuenta) in {15, 16} and tipo == TipoCuenta.card.value,
                len(cuenta) == 18 and tipo == TipoCuenta.clabe.value,
            ]
        ):
            raise ValueError('tipoCuenta no es válido')
        return tipo

    @validator('tipoCuentaBeneficiario')
    def _validate_tipoCuentaBeneficiario(cls, v, values):
        return cls._validate_tipoCuenta('cuentaBeneficiario', v, values)

    @validator('tipoCuentaOrdenante', each_item=True)
    def _validate_tipoCuentaOrdenante(cls, v, values):
        return cls._validate_tipoCuenta('cuentaOrdenante', v, values)

    @validator(
        'nombreBeneficiario', 'nombreOrdenante', 'conceptoPago', each_item=True
    )
    def _unicode_to_ascii(cls, v):
        v = unicodedata.normalize('NFKD', v).encode('ascii', 'ignore')
        return v.decode('ascii')



class StpmexException(BaseException):
    def __init__(self, **kwargs):
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    def __repr__(self):
        return (
            self.__class__.__name__
            + '('
            + ', '.join(
                [
                    f'{attr}={repr(value)}'
                    for attr, value in self.__dict__.items()
                ]
            )
            + ')'
        )

    def __str__(self):
        return repr(self)


class InvalidPassphrase(StpmexException):
    """El passphrase es incorrecto"""



from enum import Enum
from typing import Optional, Type

from pydantic import constr


def truncated_str(length) -> Type:
    return constr(strip_whitespace=True, min_length=1, curtail_length=length)


def digits(
    min_length: Optional[int] = None, max_length: Optional[int] = None
) -> Type:
    return constr(regex=r'^\d+$', min_length=min_length, max_length=max_length)


class Prioridad(Enum):
    normal = 0
    alta = 1


class TipoCuenta(Enum):
    card = 3
    phone_number = 10
    clabe = 40



import datetime as dt
import unicodedata
from typing import Any, ClassVar, Dict, Optional

from pydantic import conint, constr, validator
from pydantic.dataclasses import dataclass

from ..auth import CUENTA_FIELDNAMES, compute_signature, join_fields
from ..types import digits, truncated_str
from .base import Resource


@dataclass
class Cuenta(Resource):
    """
    Based on:
    https://stpmex.zendesk.com/hc/es/articles/360038242071-Registro-de-Cuentas-de-Personas-f%C3%ADsicas
    """

    _endpoint: ClassVar[str] = '/cuentaModule'

    nombre: truncated_str(50)
    apellidoPaterno: truncated_str(50)
    cuenta: digits(18, 18)
    rfcCurp: constr(max_length=18)

    apellidoMaterno: Optional[truncated_str(50)] = None
    genero: Optional[constr(regex=r'H|M')] = None
    fechaNacimiento: Optional[dt.date] = None
    # Esperanda para STP que agregan Nacido en Extranjero
    entidadFederativa: Optional[conint(ge=1, le=32)] = None
    actividadEconomica: Optional[conint(ge=28, le=74)] = None
    calle: Optional[truncated_str(60)] = None
    numeroExterior: Optional[digits(max_length=10)] = None
    numeroInterior: Optional[digits(max_length=5)] = None
    colonia: Optional[truncated_str(50)] = None
    alcaldiaMunicipio: Optional[truncated_str(50)] = None
    cp: Optional[digits(5, 5)] = None
    pais: Optional[conint(ge=1, lt=242)] = None
    email: Optional[constr(max_length=150)] = None
    idIdentificacion: Optional[digits(max_length=20)] = None
    telefono: Optional[digits(max_length=10)] = None

    id: Optional[int] = None

    @classmethod
    def alta(cls, **kwargs) -> 'Cuenta':
        """Dar de alta"""
        cuenta = cls(**kwargs)
        endpoint = cuenta._endpoint + '/fisica'
        resp = cuenta._client.put(endpoint, cuenta.to_dict())
        cuenta.id = resp['id']
        return cuenta

    def baja(self) -> Dict[str, Any]:
        """Dar de baja"""
        endpoint = self._endpoint + '/fisica'
        data = dict(
            cuenta=self.cuenta,
            empresa=self.empresa,
            rfcCurp=self.rfcCurp,
            firma=self.firma,
        )
        return self._client.delete(endpoint, data)

    @property
    def firma(self):
        joined_fields = join_fields(self, CUENTA_FIELDNAMES)
        return compute_signature(self._client.pkey, joined_fields)

    @validator('nombre', 'apellidoPaterno', 'apellidoMaterno', each_item=True)
    def _unicode_to_ascii(cls, v):
        v = unicodedata.normalize('NFKD', v).encode('ascii', 'ignore')
        return v.decode('ascii')



import datetime as dt
from dataclasses import asdict
from typing import Any, ClassVar, Dict


class Resource:
    _client: ClassVar['stpmex.Client']  # type: ignore
    _endpoint: ClassVar[str]
    empresa: ClassVar[str]

    @property
    def firma(self):
        ...  # pragma: no cover

    def to_dict(self) -> Dict[str, Any]:
        base = dict()
        for k, v in asdict(self).items():
            if isinstance(v, dt.date):
                base[k] = v.strftime('%Y%m%d')
            elif v:
                base[k] = v
        return {**base, **dict(firma=self.firma, empresa=self.empresa)}



from typing import Any, ClassVar, Dict

from OpenSSL import crypto
from requests import Response, Session

from .exc import InvalidPassphrase, StpmexException
from .resources import Cuenta, Orden, Resource
from .version import __version__ as client_version

DEMO_BASE_URL = 'https://demo.stpmex.com:7024/speidemows/rest'
PROD_BASE_URL = 'https://prod.stpmex.com/speiws/rest'


class Client:

    base_url: str
    demo: bool
    headers: Dict[str, str]
    session: Session

    # resources
    cuentas: ClassVar = Cuenta
    ordenes: ClassVar = Orden

    def __init__(
        self,
        empresa: str,
        priv_key: str,
        priv_key_passphrase: str,
        demo: bool = False,
    ):
        self.session = Session()
        self.headers = {'User-Agent': f'stpmex-python/{client_version}'}
        if demo:
            self.base_url = DEMO_BASE_URL
        else:
            self.base_url = PROD_BASE_URL
        try:
            self.pkey = crypto.load_privatekey(
                crypto.FILETYPE_PEM,
                priv_key,
                priv_key_passphrase.encode('ascii'),
            )
        except crypto.Error:
            raise InvalidPassphrase
        Resource.empresa = empresa
        Resource._client = self

    def put(self, endpoint: str, data: Dict[str, Any]) -> Dict[str, Any]:
        return self.request('put', endpoint, data)

    def delete(self, endpoint: str, data: Dict[str, Any]) -> Dict[str, Any]:
        return self.request('delete', endpoint, data)

    def request(
        self, method: str, endpoint: str, data: Dict[str, Any], **kwargs: Any
    ) -> Dict[str, Any]:
        url = self.base_url + endpoint
        response = self.session.request(
            method, url, json=data, headers=self.headers, **kwargs
        )
        self._check_response(response)
        resultado = response.json()
        if 'resultado' in resultado:  # Some responses are enveloped
            resultado = resultado['resultado']
        return resultado

    @staticmethod
    def _check_response(response: Response) -> None:
        if response.ok:
            resp = response.json()
            try:
                if 'descripcionError' in resp['resultado']:
                    raise StpmexException(**resp['resultado'])
            except KeyError:
                if 'descripcion' in resp and resp['descripcion']:
                    raise StpmexException(**resp)
        response.raise_for_status()
